from __future__ import with_statement

import os

from .asset_attributes import AssetAttributes
from .assets import build_asset
from .cache import Cache
from .compilers import (
    CoffeeScriptCompiler, HandlebarsCompiler, LessCompiler, StylusCompiler)
from .exceptions import FileNotFound
from .processors import DirectivesProcessor


class Finders(list):

    def register(self, finder):
        if finder not in self:
            self.append(finder)

    def unregister(self, finder):
        if finder in self:
            self.remove(finder)


class MIMETypes(dict):

    def register_defaults(self):
        self.register('.css', 'text/css')
        self.register('.js', 'application/javascript')

    def register(self, extension, mimetype):
        self[extension] = mimetype

    def unregister(self, extension):
        if extension in self:
            del self[extension]


class Compilers(dict):

    def __init__(self):
        super(Compilers, self).__init__()

    def register_defaults(self):
        self.register('.coffee', CoffeeScriptCompiler.as_handler())
        self.register('.handlebars', HandlebarsCompiler.as_handler())
        self.register('.less', LessCompiler.as_handler())
        self.register('.styl', StylusCompiler.as_handler())

    def register(self, extension, compiler):
        self[extension] = compiler

    def unregister(self, extension):
        if extension in self:
            del self[extension]


class Processors(dict):

    def register(self, mimetype, processor_class):
        if mimetype not in self or processor_class not in self[mimetype]:
            self.setdefault(mimetype, []).append(processor_class)

    def unregister(self, mimetype, processor_class):
        if mimetype in self and processor_class in self[mimetype]:
            self[mimetype].remove(processor_class)

    def get(self, mimetype):
        return super(Processors, self).get(mimetype, [])


class Preprocessors(Processors):

    def register_defaults(self):
        self.register('text/css', DirectivesProcessor.as_handler())
        self.register('application/javascript', DirectivesProcessor.as_handler())


class Postprocessors(Processors):
    pass


class Compressors(dict):

    def register(self, mimetype, compressor):
        self[mimetype] = compressor

    def unregister(self, mimetype):
        if mimetype in self:
            del self[mimetype]


class PublicAssets(list):

    def register_defaults(self):
        self.register('css/style.css')
        self.register('js/script.js')

    def register(self, path):
        if path not in self:
            self.append(path)

    def unregister(self, path):
        if path in self:
            self.remove(path)


class Suffixes(list):

    def register(self, extension, root=False, to=None, mimetype=None):
        if root:
            self.append({'extensions': [extension], 'mimetype': mimetype})
            return
        new = []
        for suffix in self:
            if to is not None and suffix['mimetype'] != to:
                continue
            extensions = list(suffix['extensions'])
            extensions.append(extension)
            new.append({'extensions': extensions, 'mimetype': mimetype})
        self.extend(new)

    def unregister(self, extension):
        for suffix in list(self):
            if extension in suffix['extensions']:
                self.remove(suffix)

    def find(self, *extensions):
        extensions = list(extensions)
        length = len(extensions)
        return [''.join(suffix['extensions']) for suffix in self
                if suffix['extensions'][:length] == extensions]


class Environment(object):

    def __init__(self, root):
        self.root = root
        self.cache = Cache()
        self.finders = Finders()
        self.compilers = Compilers()
        self.mimetypes = MIMETypes()
        self.compressors = Compressors()
        self.public_assets = PublicAssets()
        self.preprocessors = Preprocessors()
        self.postprocessors = Postprocessors()

    @property
    def suffixes(self):
        if not hasattr(self, '_suffixes'):
            suffixes = Suffixes()
            for extension, mimetype in self.mimetypes.items():
                suffixes.register(extension, root=True, mimetype=mimetype)
            for extension, compiler in self.compilers.items():
                suffixes.register(extension, to=compiler.result_mimetype)
            self._suffixes = suffixes
        return self._suffixes

    def register_defaults(self):
        self.compilers.register_defaults()
        self.mimetypes.register_defaults()
        self.public_assets.register_defaults()
        self.preprocessors.register_defaults()

    def find(self, item, logical=False):
        if isinstance(item, AssetAttributes):
            for path in item.search_paths:
                try:
                    return self.find(path, logical)
                except FileNotFound:
                    continue
            raise FileNotFound(item.path)
        if isinstance(item, (list, tuple)):
            for path in item:
                try:
                    return self.find(path, logical)
                except FileNotFound:
                    continue
        elif logical:
            asset_attribute = AssetAttributes(self, item)
            path = asset_attribute.path_without_suffix
            for suffix in self.suffixes.find(*asset_attribute.suffix):
                try:
                    return self.find(path + suffix)
                except FileNotFound:
                    continue
        else:
            for finder in self.finders:
                try:
                    absolute_path = finder.find(item)
                except FileNotFound:
                    continue
                return AssetAttributes(self, item), absolute_path
        raise FileNotFound(item)

    def list(self, path, suffix=None):
        found = set()
        suffixes = self.suffixes.find(*suffix)
        for finder in self.finders:
            for logical_path, absolute_path in finder.list(path):
                asset_attributes = AssetAttributes(self, logical_path)
                if ''.join(asset_attributes.suffix) not in suffixes:
                    continue
                if logical_path not in found:
                    yield asset_attributes, absolute_path
                    found.add(logical_path)

    def save(self):
        for path in self.public_assets:
            try:
                self.save_file(path, str(build_asset(self, path)))
            except FileNotFound:
                pass

    def save_file(self, path, source):
        filename = os.path.join(self.root, path)
        path = os.path.dirname(filename)
        if not os.path.exists(path):
            os.makedirs(path)
        elif not os.path.isdir(path):
            raise OSError("%s exists and is not a directory." % path)
        with open(filename, 'w') as f:
            f.write(source)



# -*- coding: utf-8 -*-

import os
import re
from .utils import cached_property


class AssetAttributes(object):
    """Provides access to asset path properties. The attributes object is
    created with environment object and relative (or logical) asset path.

    Some properties may be useful or not, depending on the type of passed path.
    If it is a relative asset path, you can use all properties except
    :attr:`search_paths`. In case of a logical asset path it makes sense to
    use only those properties that are not related to processors and compressor.

    :param environment: an instance of :class:`~gears.environment.Environment`
                        class.
    :param path: a relative or logical path of the asset.
    """

    def __init__(self, environment, path):
        #: Used to access the registries of compilers, processors, etc.
        #: It can be also used by asset. See
        #: :class:`~gears.environment.Environment` for more information.
        self.environment = environment

        #: The relative (or logical) path to asset.
        self.path = path

        #: The relative path to the directory the asset.
        self.dirname = os.path.dirname(path)

    @cached_property
    def search_paths(self):
        """The list of logical paths which are used to search for an asset.
        This property makes sense only if the attributes was created with
        logical path.

        It is assumed that the logical path can be a directory containing a
        file named ``index`` with the same suffix.

        Example::

            >>> attrs = AssetAttributes(environment, 'js/app.js')
            >>> attrs.search_paths
            ['js/app.js', 'js/app/index.js']

            >>> attrs = AssetAttributes(environment, 'js/app/index.js')
            >>> attrs.search_paths
            ['js/models/index.js']
        """
        paths = [self.path]
        if os.path.basename(self.path_without_suffix) != 'index':
            path = os.path.join(self.path_without_suffix, 'index')
            paths.append(path + ''.join(self.suffix))
        return paths

    @cached_property
    def path_without_suffix(self):
        """The relative path to asset without suffix.
        Example::

            >>> attrs = AssetAttributes(environment, 'js/app.js')
            >>> attrs.path_without_suffix
            'js/app'
        """
        if self.suffix:
            return self.path[:-len(''.join(self.suffix))]
        return self.path

    @property
    def logical_path(self):
        """The logical path to asset.
        Example::

            >>> attrs = AssetAttributes(environment, 'js/models.js.coffee')
            >>> attrs.logical_path
            'js/models.js'
        """
        return self.path_without_suffix + self.format_extension

    @cached_property
    def extensions(self):
        """The list of asset extensions.
        Example::

            >>> attrs = AssetAttributes(environment, 'js/models.js.coffee')
            >>> attrs.extensions
            ['.js', '.coffee']

            >>> attrs = AssetAttributes(environment, 'js/lib/external.min.js.coffee')
            >>> attrs.format_extension
            ['.min', '.js', '.coffee']
        """
        return re.findall(r'\.[^.]+', os.path.basename(self.path))

    @cached_property
    def format_extension(self):
        """The format extension of asset.
        Example::

            >>> attrs = AssetAttributes(environment, 'js/models.js.coffee')
            >>> attrs.format_extension
            '.js'

            >>> attrs = AssetAttributes(environment, 'js/lib/external.min.js.coffee')
            >>> attrs.format_extension
            '.js'
        """
        for extension in reversed(self.extensions):
            compiler = self.environment.compilers.get(extension)
            if not compiler and self.environment.mimetypes.get(extension):
                return extension

    @cached_property
    def suffix(self):
        """The list of asset extensions starting from the format extension.
        Example::

            >>> attrs = AssetAttributes(environment, 'js/lib/external.min.js.coffee')
            >>> attrs.suffix
            ['.js', '.coffee']
        """
        try:
            index = self.extensions.index(self.format_extension)
        except ValueError:
            return self.extensions
        return self.extensions[index:]

    @cached_property
    def compiler_extensions(self):
        """The list of compiler extensions.
        Example::

            >>> attrs = AssetAttributes(environment, 'js/lib/external.min.js.coffee')
            >>> attrs.suffix
            ['.coffee']
        """
        return [e for e in self.suffix[1:] if self.environment.compilers.get(e)]

    @cached_property
    def compilers(self):
        """The list of compilers used to build asset."""
        return [self.environment.compilers.get(e) for e in self.compiler_extensions]

    @property
    def preprocessors(self):
        """The list of preprocessors used to build asset."""
        return self.environment.preprocessors.get(self.mimetype)

    @property
    def postprocessors(self):
        """The list of postprocessors used to build asset."""
        return self.environment.postprocessors.get(self.mimetype)

    @property
    def processors(self):
        """The list of all processors (preprocessors, compilers,
        postprocessors) used to build asset.
        """
        return self.preprocessors + list(reversed(self.compilers)) + self.postprocessors

    @property
    def compressor(self):
        """The compressors used to compress the asset."""
        return self.environment.compressors.get(self.mimetype)

    @cached_property
    def mimetype(self):
        """MIME-type of the asset."""
        return (self.environment.mimetypes.get(self.format_extension) or
                'application/octet-stream')



from gears.environment import Suffixes
from unittest2 import TestCase


class SuffixesTests(TestCase):

    def setUp(self):
        self.suffixes = Suffixes()

    def test_register_root_extensions(self):
        self.suffixes.register('.css', root=True, mimetype='text/css')
        self.suffixes.register('.txt', root=True, mimetype='text/plain')
        self.assertEqual(
            self.suffixes,
            [{'extensions': ['.css'], 'mimetype': 'text/css'},
             {'extensions': ['.txt'], 'mimetype': 'text/plain'}])

    def test_register_extension_to_mimetype(self):
        self.suffixes.register('.css', root=True, mimetype='text/css')
        self.suffixes.register('.txt', root=True, mimetype='text/plain')
        self.suffixes.register('.styl', to='text/css')
        self.assertEqual(
            self.suffixes,
            [{'extensions': ['.css'], 'mimetype': 'text/css'},
             {'extensions': ['.txt'], 'mimetype': 'text/plain'},
             {'extensions': ['.css', '.styl'], 'mimetype': None}])

    def test_register_extension_to_none(self):
        self.suffixes.register('.css', root=True, mimetype='text/css')
        self.suffixes.register('.txt', root=True, mimetype='text/plain')
        self.suffixes.register('.styl', to='text/css')
        self.suffixes.register('.tmpl')
        self.assertEqual(
            self.suffixes,
            [{'extensions': ['.css'], 'mimetype': 'text/css'},
             {'extensions': ['.txt'], 'mimetype': 'text/plain'},
             {'extensions': ['.css', '.styl'], 'mimetype': None},
             {'extensions': ['.css', '.tmpl'], 'mimetype': None},
             {'extensions': ['.txt', '.tmpl'], 'mimetype': None},
             {'extensions': ['.css', '.styl', '.tmpl'], 'mimetype': None}])

    def test_unregister_extension(self):
        self.suffixes.register('.css', root=True, mimetype='text/css')
        self.suffixes.register('.txt', root=True, mimetype='text/plain')
        self.suffixes.register('.styl', to='text/css')
        self.suffixes.register('.tmpl')
        self.suffixes.unregister('.css')
        self.assertEqual(
            self.suffixes,
            [{'extensions': ['.txt'], 'mimetype': 'text/plain'},
             {'extensions': ['.txt', '.tmpl'], 'mimetype': None}])

    def test_find_all(self):
        self.suffixes.register('.css', root=True, mimetype='text/css')
        self.suffixes.register('.txt', root=True, mimetype='text/plain')
        self.suffixes.register('.styl', to='text/css')
        self.assertEqual(self.suffixes.find(), ['.css', '.txt', '.css.styl'])

    def test_find_by_extension(self):
        self.suffixes.register('.css', root=True, mimetype='text/css')
        self.suffixes.register('.txt', root=True, mimetype='text/plain')
        self.suffixes.register('.styl', to='text/css')
        self.assertEqual(self.suffixes.find('.css'), ['.css', '.css.styl'])

    def test_find_by_many_extensions(self):
        self.suffixes.register('.css', root=True, mimetype='text/css')
        self.suffixes.register('.txt', root=True, mimetype='text/plain')
        self.suffixes.register('.styl', to='text/css')
        self.assertEqual(self.suffixes.find('.css', '.styl'), ['.css.styl'])

    def test_find_nothing(self):
        self.assertEqual(self.suffixes.find('.css'), [])



from gears.environment import Processors, Preprocessors
from gears.processors import DirectivesProcessor

from mock import Mock
from unittest2 import TestCase


class ProcessorsTests(TestCase):

    def setUp(self):
        self.processors = Processors()
        self.first_processor = Mock()
        self.second_processor = Mock()
        self.third_processor = Mock()

    def test_register(self):
        self.processors.register('text/css', self.first_processor)
        self.processors.register('text/css', self.second_processor)
        self.processors.register('text/plain', self.third_processor)
        self.assertIn('text/css', self.processors)
        self.assertEqual(self.processors['text/css'], [self.first_processor, self.second_processor])
        self.assertIn('text/plain', self.processors)
        self.assertEqual(self.processors['text/plain'], [self.third_processor])

    def test_register_twice(self):
        self.processors.register('text/css', self.first_processor)
        self.processors.register('text/css', self.first_processor)
        self.assertEqual(self.processors['text/css'], [self.first_processor])

    def test_unregister(self):
        self.processors.register('text/css', self.first_processor)
        self.processors.register('text/css', self.second_processor)
        self.processors.unregister('text/css', self.first_processor)
        self.assertEqual(self.processors['text/css'], [self.second_processor])

    def test_unregister_if_does_not_exist(self):
        self.processors.unregister('text/css', self.first_processor)

    def test_get_if_exists(self):
        self.processors.register('text/css', self.first_processor)
        self.processors.register('text/css', self.second_processor)
        self.assertEqual(self.processors.get('text/css'), [self.first_processor, self.second_processor])

    def test_get_if_does_not_exist(self):
        self.assertEqual(self.processors.get('text/css'), [])


class PreprocessorsTests(TestCase):

    def setUp(self):
        self.preprocessors = Preprocessors()

    def test_register_defaults(self):
        self.preprocessors.register_defaults()
        self.assertItemsEqual(self.preprocessors.keys(), ['text/css', 'application/javascript'])
        self.assertEqual(
            [p.handler_class for p in self.preprocessors['text/css']],
            [DirectivesProcessor])
        self.assertEqual(
            [p.handler_class for p in self.preprocessors['application/javascript']],
            [DirectivesProcessor])



from gears.environment import Finders
from mock import Mock
from unittest2 import TestCase


class FindersTests(TestCase):

    def setUp(self):
        self.finders = Finders()
        self.first_finder = Mock()
        self.second_finder = Mock()

    def test_register(self):
        self.finders.register(self.first_finder)
        self.finders.register(self.second_finder)
        self.assertEqual(self.finders, [self.first_finder, self.second_finder])

    def test_register_twice(self):
        self.finders.register(self.first_finder)
        self.finders.register(self.first_finder)
        self.assertEqual(self.finders, [self.first_finder])

    def test_unregister(self):
        self.finders.register(self.first_finder)
        self.finders.unregister(self.first_finder)
        self.assertEqual(self.finders, [])



from gears.compilers import (
    CoffeeScriptCompiler, HandlebarsCompiler, LessCompiler, StylusCompiler)
from gears.environment import Compilers

from mock import Mock
from unittest2 import TestCase


class CompilersTests(TestCase):

    def setUp(self):
        self.compilers = Compilers()
        self.first_compiler = Mock()
        self.second_compiler = Mock()

    def test_register_defaults(self):
        self.compilers.register_defaults()
        self.assertItemsEqual(self.compilers, ['.coffee', '.handlebars', '.less', '.styl'])
        self.assertIs(self.compilers['.coffee'].handler_class, CoffeeScriptCompiler)
        self.assertIs(self.compilers['.handlebars'].handler_class, HandlebarsCompiler)
        self.assertIs(self.compilers['.less'].handler_class, LessCompiler)
        self.assertIs(self.compilers['.styl'].handler_class, StylusCompiler)

    def test_register(self):
        self.compilers.register('.css', self.first_compiler)
        self.assertIn('.css', self.compilers)
        self.assertIs(self.compilers['.css'], self.first_compiler)

    def test_register_twice(self):
        self.compilers.register('.css', self.first_compiler)
        self.compilers.register('.css', self.second_compiler)
        self.assertIs(self.compilers['.css'], self.second_compiler)

    def test_unregister(self):
        self.compilers.register('.css', self.first_compiler)
        self.compilers.unregister('.css')
        self.assertNotIn('.css', self.compilers)

    def test_unregister_if_does_not_exist(self):
        self.compilers.unregister('.css')



from gears.assets import Requirements
from unittest2 import TestCase


class FakeAsset(object):

    def __init__(self):
        self.requirements = Requirements(self)


class AssetRequirementsTests(TestCase):

    def setUp(self):
        self.asset = FakeAsset()

        self.a = FakeAsset()
        self.b = FakeAsset()
        self.c = FakeAsset()
        self.d = FakeAsset()
        self.e = FakeAsset()

        self.asset.requirements.add(self.a)
        self.asset.requirements.add(self.b)
        self.asset.requirements.add(self.c)
        self.asset.requirements.add(self.asset)
        self.asset.requirements.add(self.d)
        self.asset.requirements.add(self.e)

    def test_firstly_adds_before(self):
        self.assertEqual(self.asset.requirements.before, [self.a, self.b, self.c])

    def test_adds_after_in_the_end(self):
        self.assertEqual(self.asset.requirements.after, [self.d, self.e])



from __future__ import with_statement

import os

from gears.assets import Requirements, Asset
from gears.environment import Environment
from gears.finders import FileSystemFinder
from gears.processors import DirectivesProcessor

from unittest2 import TestCase


TESTS_DIR = os.path.dirname(__file__)
FIXTURES_DIR = os.path.join(TESTS_DIR, 'fixtures', 'directives_processor')


class DirectivesProcessorTests(TestCase):

    def check_paths(self, assets, paths):
        self.assertEqual([asset.attributes.path for asset in assets], paths)

    def get_fixture_path(self, fixture):
        return os.path.join(FIXTURES_DIR, fixture)

    def get_finder(self, fixture):
        return FileSystemFinder([self.get_fixture_path(fixture)])

    def get_environment(self, fixture):
        environment = Environment(os.path.join(TESTS_DIR, 'static'))
        environment.finders.register(self.get_finder(fixture))
        environment.mimetypes.register_defaults()
        environment.preprocessors.register_defaults()
        return environment

    def get_asset(self, fixture):
        return Asset(*self.get_environment(fixture).find('source.js'))

    def get_output_source(self, fixture):
        fixture_path = self.get_fixture_path(fixture)
        with open(os.path.join(fixture_path, 'output.js')) as f:
            return f.read()

    def test_fills_asset_requirements(self):
        asset = self.get_asset('requirements')
        DirectivesProcessor.as_handler()(asset)
        self.check_paths(asset.requirements.before, ['js/external.js'])
        self.check_paths(asset.requirements.after, ['js/models.js', 'js/views.js'])

    def test_modifies_processed_source(self):
        asset = self.get_asset('requirements')
        DirectivesProcessor.as_handler()(asset)
        self.assertEqual(asset.processed_source, self.get_output_source('requirements'))

    def test_requires_asset_only_once(self):
        asset = self.get_asset('require_multiple_times')
        DirectivesProcessor.as_handler()(asset)
        self.check_paths(list(asset.requirements), 'external.js models.js views.js source.js'.split())



from __future__ import with_statement
from gears.utils import safe_join
from unittest2 import TestCase


class SafeJoinTests(TestCase):

    def test_if_base_is_not_absolute(self):
        with self.assertRaisesRegexp(ValueError, 'is not an absolute path'):
            safe_join('assets', 'js/script.js')

    def test_if_path_is_outside_of_base(self):
        with self.assertRaisesRegexp(ValueError, 'is outside of'):
            safe_join('/assets', '../js/script.js')

    def test_success(self):
        path = safe_join('/assets', 'js/script.js')
        self.assertEqual(path, '/assets/js/script.js')



import os
from .exceptions import ImproperlyConfigured, FileNotFound
from .utils import safe_join


class BaseFinder(object):

    def find(self, path):
        raise NotImplementedError()


class FileSystemFinder(BaseFinder):

    def __init__(self, directories):
        self.locations = []
        if not isinstance(directories, (list, tuple)):
            raise ImproperlyConfigured(
                "FileSystemFinder's 'directories' parameter is not a "
                "tuple or list; perhaps you forgot a trailing comma?")
        for directory in directories:
            if directory not in self.locations:
                self.locations.append(directory)

    def find(self, path):
        for matched_path in self.find_all(path):
            return matched_path
        raise FileNotFound(path)

    def find_all(self, path):
        for root in self.locations:
            matched_path = self.find_location(root, path)
            if matched_path:
                yield matched_path

    def find_location(self, root, path):
        path = safe_join(root, path)
        if os.path.exists(path):
            return path

    def list(self, path):
        for root in self.locations:
            matched_path = self.find_location(root, path)
            if not matched_path or not os.path.isdir(matched_path):
                continue
            for filename in os.listdir(matched_path):
                filepath = os.path.join(matched_path, filename)
                if os.path.isfile(filepath):
                    yield os.path.join(path, filename), filepath



# -*- coding: utf-8 -*-

from functools import wraps


class BaseAssetHandler(object):
    """Base class for all asset handlers (processors, compilers and
    compressors). A subclass has to implement :meth:`__call__` which is called
    with asset as argument.
    """

    def __call__(self, asset):
        """Subclasses have to override this method to implement the actual
        handler function code. This method is called with asset as argument.
        Depending on the type of the handler, this method must change asset
        state (as it does in :class:`~gears.processors.Directivesprocessor`)
        or return some value (in case of asset compressors).
        """
        raise NotImplementedError

    @classmethod
    def as_handler(cls, **initkwargs):
        """Converts the class into an actual handler function that can be used
        when registering different types of processors in
        :class:`~gears.enviroment.Environment` class instance.

        The arguments passed to :meth:`as_handler` are forwarded to the
        constructor of the class.
        """
        @wraps(cls, updated=())
        def handler(asset):
            return handler.handler_class(**initkwargs)(asset)
        handler.handler_class = cls
        return handler



from __future__ import with_statement

import os
import shutil
from contextlib import contextmanager

from gears.asset_attributes import AssetAttributes
from gears.environment import Environment
from gears.exceptions import FileNotFound
from gears.finders import FileSystemFinder

from mock import Mock
from unittest2 import TestCase


ASSETS_DIR = os.path.abspath(os.path.join(os.path.dirname(__file__), 'assets'))
STATIC_DIR = os.path.abspath(os.path.join(os.path.dirname(__file__), 'static'))


class FakeCompiler(object):

    def __init__(self, result_mimetype):
        self.result_mimetype = result_mimetype


class FakeFinder(object):

    paths = ('js/script.js', 'js/app/index.js', 'js/models.js.coffee')

    def find(self, path):
        if path in self.paths:
            return '/assets/' + path
        raise FileNotFound(path)


@contextmanager
def remove_static_dir():
    if os.path.exists(STATIC_DIR):
        shutil.rmtree(STATIC_DIR)
    yield
    if os.path.exists(STATIC_DIR):
        shutil.rmtree(STATIC_DIR)


class EnvironmentTests(TestCase):

    def setUp(self):
        self.environment = Environment(STATIC_DIR)

    def test_suffixes(self):
        self.environment.mimetypes.register('.css', 'text/css')
        self.environment.mimetypes.register('.txt', 'text/plain')
        self.environment.compilers.register('.styl', FakeCompiler('text/css'))
        self.assertItemsEqual(self.environment.suffixes.find(),
                              ['.css', '.css.styl', '.txt'])

    def test_register_defaults(self):
        self.environment.compilers = Mock()
        self.environment.mimetypes = Mock()
        self.environment.public_assets = Mock()
        self.environment.preprocessors = Mock()
        self.environment.register_defaults()
        self.environment.compilers.register_defaults.assert_called_once_with()
        self.environment.mimetypes.register_defaults.assert_called_once_with()
        self.environment.public_assets.register_defaults.assert_called_once_with()
        self.environment.preprocessors.register_defaults.assert_called_once_with()


class EnvironmentFindTests(TestCase):

    def setUp(self):
        self.environment = Environment(STATIC_DIR)
        self.environment.register_defaults()
        self.environment.finders.register(FakeFinder())
        self.environment.compilers.register(
            '.coffee', FakeCompiler('application/javascript'))

    def check_asset_attributes(self, attrs, path):
        self.assertIsInstance(attrs, AssetAttributes)
        self.assertIs(attrs.environment, self.environment)
        self.assertEqual(attrs.path, path)

    def test_find_by_path(self):
        attrs, path = self.environment.find('js/models.js.coffee')
        self.check_asset_attributes(attrs, 'js/models.js.coffee')
        self.assertEqual(path, '/assets/js/models.js.coffee')

    def test_find_nothing_by_path(self):
        with self.assertRaises(FileNotFound):
            self.environment.find('js/models.js')

    def test_find_by_path_list(self):
        attrs, path = self.environment.find(['js/app.js', 'js/app/index.js'])
        self.check_asset_attributes(attrs, 'js/app/index.js')
        self.assertEqual(path, '/assets/js/app/index.js')

    def test_find_nothing_by_path_list(self):
        with self.assertRaises(FileNotFound):
            self.environment.find(['style.css', 'style/index.css'])

    def test_find_by_asset_attributes(self):
        attrs = AssetAttributes(self.environment, 'js/app.js')
        attrs, path = self.environment.find(attrs)
        self.check_asset_attributes(attrs, 'js/app/index.js')
        self.assertEqual(path, '/assets/js/app/index.js')

    def test_find_nothing_by_asset_attributes(self):
        attrs = AssetAttributes(self.environment, 'js/models.js')
        with self.assertRaises(FileNotFound):
            self.environment.find(attrs)

    def test_find_by_logical_path(self):
        attrs, path = self.environment.find('js/models.js', logical=True)
        self.check_asset_attributes(attrs, 'js/models.js.coffee')
        self.assertEqual(path, '/assets/js/models.js.coffee')

    def test_find_nothing_by_logical_path(self):
        with self.assertRaises(FileNotFound):
            self.environment.find('js/views.js', logical=True)

    def test_save_file(self):
        with remove_static_dir():
            self.environment.save_file('js/script.js', 'hello world')
            with open(os.path.join(STATIC_DIR, 'js', 'script.js')) as f:
                self.assertEqual(f.read(), 'hello world')


class EnvironmentListTests(TestCase):

    def setUp(self):
        self.environment = Environment(STATIC_DIR)
        self.environment.register_defaults()
        self.environment.finders.register(FileSystemFinder([ASSETS_DIR]))

    def test_list(self):
        items = list(self.environment.list('js/templates', ['.js', '.handlebars']))
        self.assertEqual(len(items), 3)
        for i, item in enumerate(sorted(items, key=lambda x: x[1])):
            path = 'js/templates/%s.js.handlebars' % 'abc'[i]
            asset_attributes, absolute_path = item
            self.assertIsInstance(asset_attributes, AssetAttributes)
            self.assertEqual(asset_attributes.path, path)
            self.assertEqual(absolute_path, os.path.join(ASSETS_DIR, path))



from __future__ import with_statement

import os

from gears.exceptions import ImproperlyConfigured, FileNotFound
from gears.finders import FileSystemFinder

from mock import patch, Mock
from unittest2 import TestCase


ASSETS_DIR = os.path.abspath(os.path.join(os.path.dirname(__file__), 'assets'))


class FileSystemFinderTests(TestCase):

    def test_initialization(self):
        finder = FileSystemFinder(('/first', '/second', '/third', '/first'))
        self.assertEqual(finder.locations, ['/first', '/second', '/third'])

    def test_if_directories_is_not_iterable(self):
        with self.assertRaises(ImproperlyConfigured):
            finder = FileSystemFinder('/first')

    @patch('os.path.exists')
    def test_find_location_if_exists(self, exists):
        exists.return_value = True
        finder = FileSystemFinder(('/assets',))
        location = finder.find_location('/assets', 'js/script.js')
        self.assertEqual(location, '/assets/js/script.js')
        exists.assert_called_once_with('/assets/js/script.js')

    @patch('os.path.exists')
    def test_find_location_if_does_not_exist(self, exists):
        exists.return_value = False
        finder = FileSystemFinder(('/assets',))
        self.assertIsNone(finder.find_location('/assets', 'js/script.js'))
        exists.assert_called_once_with('/assets/js/script.js')

    def test_find_if_exists(self):
        finder = FileSystemFinder(('/first', '/second', '/third'))
        finder.find_all = Mock(return_value=(
            '/second/js/script.js', '/third/js/script.js'))
        self.assertEqual(finder.find('js/script.js'), '/second/js/script.js')
        finder.find_all.assert_called_once_with('js/script.js')

    def test_find_if_does_not_exist(self):
        finder = FileSystemFinder(('/first', '/second', '/third'))
        finder.find_all = Mock(return_value=())
        with self.assertRaises(FileNotFound):
            finder.find('js/script.js')
        finder.find_all.assert_called_once_with('js/script.js')

    def test_find_all_if_exists(self):

        def find_location(root, path):
            if root != '/first':
                return os.path.join(root, path)

        finder = FileSystemFinder(('/first', '/second', '/third'))
        finder.find_location = Mock(side_effect=find_location)

        paths = list(finder.find_all('js/script.js'))
        self.assertEqual(paths, ['/second/js/script.js', '/third/js/script.js'])
        self.assertEqual(finder.find_location.call_args_list, [
            (('/first', 'js/script.js'), {}),
            (('/second', 'js/script.js'), {}),
            (('/third', 'js/script.js'), {})])

    def test_find_all_if_does_not_exist(self):
        finder = FileSystemFinder(('/first', '/second', '/third'))
        finder.find_location = Mock(return_value=False)

        self.assertEqual(list(finder.find_all('js/script.js')), [])
        self.assertEqual(finder.find_location.call_args_list, [
            (('/first', 'js/script.js'), {}),
            (('/second', 'js/script.js'), {}),
            (('/third', 'js/script.js'), {})])

    def test_find(self):
        finder = FileSystemFinder([ASSETS_DIR])
        self.assertItemsEqual(finder.list('js/templates'), (
            ('js/templates/readme.txt', os.path.join(ASSETS_DIR, 'js/templates/readme.txt')),
            ('js/templates/a.js.handlebars', os.path.join(ASSETS_DIR, 'js/templates/a.js.handlebars')),
            ('js/templates/b.js.handlebars', os.path.join(ASSETS_DIR, 'js/templates/b.js.handlebars')),
            ('js/templates/c.js.handlebars', os.path.join(ASSETS_DIR, 'js/templates/c.js.handlebars')),
        ))



from gears.asset_attributes import AssetAttributes
from gears.environment import Environment

from mock import Mock
from unittest2 import TestCase


class AssetAttributesTests(TestCase):

    def setUp(self):
        self.environment = Environment('assets')

    def create_attributes(self, path):
        return AssetAttributes(self.environment, path)

    def test_extensions(self):

        def check(path, expected_result):
            extensions = self.create_attributes(path).extensions
            self.assertEqual(extensions, expected_result)

        check('js/readme', [])
        check('js/script.js', ['.js'])
        check('js/script.js.coffee', ['.js', '.coffee'])
        check('js/app.min.js.coffee', ['.min', '.js', '.coffee'])
        check('js/.htaccess', ['.htaccess'])

    def test_path_without_suffix(self):

        def check(path, expected_result):
            attributes = self.create_attributes(path)
            path_without_suffix = attributes.path_without_suffix
            self.assertEqual(path_without_suffix, expected_result)

        check('js/.htaccess', 'js/')
        check('js/readme', 'js/readme')
        check('js/app.min.js.coffee', 'js/app')

        self.environment.mimetypes.register('.js', 'application/javascript')
        check('js/app.min.js.coffee', 'js/app.min')

    def test_logical_path(self):
        self.environment.mimetypes.register('.js', 'application/javascript')

        def check(path, expected_result):
            logical_path = self.create_attributes(path).logical_path
            self.assertEqual(logical_path, expected_result)

        check('js/script.js', 'js/script.js')
        check('js/script.js.coffee', 'js/script.js')
        check('js/script.min.js', 'js/script.min.js')
        check('js/script.min.js.coffee', 'js/script.min.js')

    def test_search_paths(self):
        self.environment.mimetypes.register('.js', 'application/javascript')

        def check(path, expected_result):
            search_paths = self.create_attributes(path).search_paths
            self.assertEqual(search_paths, expected_result)

        check('js/script.js', ['js/script.js', 'js/script/index.js'])
        check('js/app.min.js', ['js/app.min.js', 'js/app.min/index.js'])
        check('js/.htaccess', ['js/.htaccess', 'js/index.htaccess'])
        check('js/index.js', ['js/index.js'])

    def test_format_extension(self):
        self.environment.mimetypes.register('.css', 'text/css')
        self.environment.mimetypes.register('.js', 'application/javascript')
        self.environment.compilers.register('.coffee', Mock())

        def check(path, expected_result):
            format_extension = self.create_attributes(path).format_extension
            self.assertEqual(format_extension, expected_result)

        check('js/script.js', '.js')
        check('js/script.js.coffee', '.js')
        check('js/jquery.min.js', '.js')
        check('js/app.js.min.coffee', '.js')
        check('css/style.js.css', '.css')
        check('readme', None)

    def test_suffix(self):
        self.environment.mimetypes.register('.css', 'text/css')
        self.environment.mimetypes.register('.js', 'application/javascript')
        self.environment.compilers.register('.coffee', Mock())

        def check(path, expected_result):
            suffix = self.create_attributes(path).suffix
            self.assertEqual(suffix, expected_result)

        check('js/script.js', ['.js'])
        check('js/script.js.coffee', ['.js', '.coffee'])
        check('js/script.min.js.coffee', ['.js', '.coffee'])
        check('js/script.js.min.coffee', ['.js', '.min', '.coffee'])
        check('readme', [])

    def test_compiler_extensions(self):
        self.environment.mimetypes.register('.css', 'text/css')
        self.environment.mimetypes.register('.js', 'application/javascript')
        self.environment.compilers.register('.coffee', Mock())
        self.environment.compilers.register('.tmpl', Mock())

        def check(path, expected_result):
            compiler_extensions = self.create_attributes(path).compiler_extensions
            self.assertEqual(compiler_extensions, expected_result)

        check('js/script.js', [])
        check('js/script.js.coffee', ['.coffee'])
        check('js/script.js.coffee.tmpl', ['.coffee', '.tmpl'])
        check('js/hot.coffee.js.tmpl', ['.tmpl'])

    def test_compilers(self):
        coffee_compiler = Mock()
        template_compiler = Mock()
        self.environment.mimetypes.register('.css', 'text/css')
        self.environment.mimetypes.register('.js', 'application/javascript')
        self.environment.compilers.register('.coffee', coffee_compiler)
        self.environment.compilers.register('.tmpl', template_compiler)

        def check(path, expected_result):
            compilers = self.create_attributes(path).compilers
            self.assertEqual(compilers, expected_result)

        check('js/script.js', [])
        check('js/script.js.coffee', [coffee_compiler])
        check('js/script.js.coffee.tmpl', [coffee_compiler, template_compiler])
        check('js/hot.coffee.js.tmpl', [template_compiler])

    def test_mimetype(self):
        self.environment.mimetypes.register('.css', 'text/css')
        self.environment.mimetypes.register('.js', 'application/javascript')
        self.environment.compilers.register('.coffee', Mock())
        self.environment.compilers.register('.styl', Mock())

        def check(path, expected_result):
            mimetype = self.create_attributes(path).mimetype
            self.assertEqual(mimetype, expected_result)

        check('js/script.js', 'application/javascript')
        check('js/script.js.coffee', 'application/javascript')
        check('css/style.min.css', 'text/css')
        check('readme.txt', 'application/octet-stream')

    def test_preprocessors(self):
        first_processor = Mock()
        second_processor = Mock()
        self.environment.mimetypes.register('.css', 'text/css')
        self.environment.mimetypes.register('.js', 'application/javascript')
        preprocessors = self.environment.preprocessors
        preprocessors.register('text/css', first_processor)
        preprocessors.register('text/css', second_processor)

        def check(path, expected_result):
            preprocessors = self.create_attributes(path).preprocessors
            self.assertEqual(preprocessors, expected_result)

        check('readme.txt', [])
        check('js/script.js', [])
        check('css/style.css', [first_processor, second_processor])

    def test_postprocessors(self):
        first_processor = Mock()
        second_processor = Mock()
        self.environment.mimetypes.register('.css', 'text/css')
        self.environment.mimetypes.register('.js', 'application/javascript')
        postprocessors = self.environment.postprocessors
        postprocessors.register('text/css', first_processor)
        postprocessors.register('text/css', second_processor)

        def check(path, expected_result):
            postprocessors = self.create_attributes(path).postprocessors
            self.assertEqual(postprocessors, expected_result)

        check('readme.txt', [])
        check('js/script.js', [])
        check('css/style.css', [first_processor, second_processor])



from gears.environment import MIMETypes
from unittest2 import TestCase


class MIMETypesTests(TestCase):

    def setUp(self):
        self.mimetypes = MIMETypes()
        self.mimetypes.register_defaults()

    def test_register(self):
        self.mimetypes.register('.txt', 'text/plain')
        self.assertIn('.txt', self.mimetypes)
        self.assertEqual(self.mimetypes['.txt'], 'text/plain')

    def test_unregister(self):
        self.assertIn('.css', self.mimetypes)
        self.mimetypes.unregister('.css')
        self.assertNotIn('.css', self.mimetypes)

    def test_get_if_exists(self):
        self.assertEqual(self.mimetypes.get('.css'), 'text/css')

    def test_get_if_does_not_exist(self):
        self.assertIsNone(self.mimetypes.get('.txt'))



from __future__ import with_statement

import codecs
import hashlib
import os

from .asset_attributes import AssetAttributes
from .utils import cached_property, unique


class CircularDependencyError(Exception):

    def __init__(self, absolute_path):
        self.absolute_path = absolute_path

    def __str__(self):
        return '%s has already been required' % self.absolute_path


class Requirements(object):

    def __init__(self, asset):
        self.current = self.before = []
        self.after = []
        self.asset = asset

    def __iter__(self):
        return self._iter_unique()

    def __repr__(self):
        return '<Requirements before=%r after=%r>' % (self.before, self.after)

    @classmethod
    def from_dict(cls, asset, data):
        self = cls(asset)
        for absolute_path, logical_path in data['before']:
            self.add(self._asset_from_paths(absolute_path, logical_path))
        self.add(asset)
        for absolute_path, logical_path in data['after']:
            self.add(self._asset_from_paths(absolute_path, logical_path))
        return self

    @cached_property
    def expired(self):
        return (any(asset.bundle_expired for asset in self.before) or 
                any(asset.bundle_expired for asset in self.after))

    def add(self, asset):
        if asset is self.asset:
            self.current = self.after
        else:
            self.current.append(asset)

    def to_dict(self):
        return {'before': [self._paths_from_asset(asset) for asset in self.before],
                'after': [self._paths_from_asset(asset) for asset in self.after]}

    def _iter_requirements(self, assets):
        for asset in assets:
            for requirement in asset.requirements:
                yield requirement

    def _iter_all(self):
        for requirement in self._iter_requirements(self.before):
            yield requirement
        yield self.asset
        for requirement in self._iter_requirements(self.after):
            yield requirement

    def _iter_unique(self):
        return unique(self._iter_all(), key=lambda asset: asset.absolute_path)

    def _asset_from_paths(self, absolute_path, logical_path):
        attributes = AssetAttributes(self.asset.attributes.environment, logical_path)
        return Asset(attributes, absolute_path)

    def _paths_from_asset(self, asset):
        return (asset.absolute_path, asset.attributes.path)


class BaseAsset(object):

    def __init__(self, attributes, absolute_path, calls=None):
        self.attributes = attributes
        self.absolute_path = absolute_path

        self.calls = calls.copy() if calls else set()
        if self.absolute_path in self.calls:
            raise CircularDependencyError(self.absolute_path)
        self.calls.add(self.absolute_path)

    def __str__(self):
        return self.source

    def __repr__(self):
        return '<%s absolute_path=%s>' % (self.__class__.__name__, self.absolute_path)


class Asset(BaseAsset):

    def __init__(self, *args, **kwargs):
        super(Asset, self).__init__(*args, **kwargs)
        if self.expired:
            self.requirements = Requirements(self)
            self.processed_source = self.source
            for process in self.attributes.processors:
                process(self)
            self.attributes.environment.cache.set(self)
        else:
            self._init_from_cache()

    def __unicode__(self):
        return self.compressed_source

    def __str__(self):
        return unicode(self).encode('utf-8')

    @cached_property
    def source(self):
        with codecs.open(self.absolute_path, encoding='utf-8') as f:
            return f.read()

    @cached_property
    def bundled_source(self):
        data = self.attributes.environment.cache.get(self)
        if not self.bundle_expired and 'bundled_source' in data:
            return data['bundled_source']
        bundled_source = u'\n'.join(r.processed_source for r in self.requirements)
        data['bundled_source'] = bundled_source
        return bundled_source

    @cached_property
    def compressed_source(self):
        data = self.attributes.environment.cache.get(self)
        if not self.bundle_expired and 'compressed_source' in data:
            return data['compressed_source']
        compressed_source = self.bundled_source
        compress = self.attributes.compressor
        if compress:
            compressed_source = compress(self)
        data['compressed_source'] = compressed_source
        return compressed_source

    @cached_property
    def mtime(self):
        return os.stat(self.absolute_path).st_mtime

    @cached_property
    def hexdigest(self):
        return hashlib.sha1(self.source.encode('utf-8')).hexdigest()

    @cached_property
    def expired(self):
        data = self.attributes.environment.cache.get(self)
        return (data is None or
                self.mtime > data['mtime'] or
                self.hexdigest > data['hexdigest'])

    @cached_property
    def bundle_expired(self):
        return self.expired or self.requirements.expired

    def to_dict(self):
        return {'processed_source': self.processed_source,
                'requirements': self.requirements.to_dict(),
                'hexdigest': self.hexdigest,
                'mtime': self.mtime}

    def _init_from_cache(self):
        data = self.attributes.environment.cache.get(self)
        self.requirements = Requirements.from_dict(self, data['requirements'])
        self.processed_source = data['processed_source']



class StaticAsset(BaseAsset):

    @cached_property
    def source(self):
        with open(self.absolute_path, 'rb') as f:
            return f.read()


def build_asset(environment, path):
    asset_attributes = AssetAttributes(environment, path)
    asset_attributes, absolute_path = environment.find(asset_attributes, True)
    if not asset_attributes.processors:
        return StaticAsset(asset_attributes, absolute_path)
    return Asset(asset_attributes, absolute_path)



import os
import shlex
from functools import wraps

from .base import BaseProcessor
from ..asset_attributes import AssetAttributes
from ..assets import Asset
from ..directives_parser import DirectivesParser


class DirectivesProcessor(BaseProcessor):

    def __init__(self):
        self.types = {
            'require': self.process_require_directive,
            'require_directory': self.process_require_directory_directive,
            'require_self': self.process_require_self_directive,
        }

    def __call__(self, asset):
        self.asset = asset
        self.parse()
        self.process_directives()

    def parse(self):
        directives, source = DirectivesParser().parse(self.asset.processed_source)
        self.directives = directives
        self.asset.processed_source = source

    def process_directives(self):
        for directive in self.directives:
            args = shlex.split(directive.encode('utf-8'))
            self.types[args[0]](*args[1:])

    def process_require_directive(self, path):
        self.asset.requirements.add(self.get_asset(*self.find(path)))

    def process_require_directory_directive(self, path):
        path = self.get_relative_path(path, is_directory=True)
        list = self.asset.attributes.environment.list(path, self.asset.attributes.suffix)
        for asset_attributes, absolute_path in sorted(list, key=lambda x: x[0].path):
            self.asset.requirements.add(self.get_asset(asset_attributes, absolute_path))

    def process_require_self_directive(self):
        self.asset.requirements.add(self.asset)

    def find(self, require_path):
        require_path = self.get_relative_path(require_path)
        asset_attributes = AssetAttributes(self.asset.attributes.environment, require_path)
        return self.asset.attributes.environment.find(asset_attributes, True)

    def get_relative_path(self, require_path, is_directory=False):
        require_path = os.path.join(self.asset.attributes.dirname, require_path)
        require_path = os.path.normpath(require_path)
        if is_directory:
            return require_path
        return require_path + ''.join(self.asset.attributes.extensions)

    def get_asset(self, asset_attributes, absolute_path):
        return Asset(asset_attributes, absolute_path, self.asset.calls)



from __future__ import with_statement

import codecs
import os

from gears.assets import CircularDependencyError, Asset
from gears.environment import Environment
from gears.finders import FileSystemFinder
from gears.processors import DirectivesProcessor

from unittest2 import TestCase


TESTS_DIR = os.path.dirname(__file__)
FIXTURES_DIR = os.path.join(TESTS_DIR, 'fixtures', 'assets')


def read(file):
    with codecs.open(file, encoding='utf-8') as f:
        return f.read()


class AssetTests(TestCase):

    def get_fixture_path(self, fixture):
        return os.path.join(FIXTURES_DIR, fixture)

    def get_finder(self, fixture):
        return FileSystemFinder([self.get_fixture_path(fixture)])

    def get_environment(self, fixture):
        environment = Environment(os.path.join(TESTS_DIR, 'static'))
        environment.finders.register(self.get_finder(fixture))
        environment.mimetypes.register_defaults()
        environment.preprocessors.register_defaults()
        return environment

    def get_asset(self, fixture):
        return Asset(*self.get_environment(fixture).find('source.js'))

    def test_circular_dependency(self):
        with self.assertRaises(CircularDependencyError):
            asset = self.get_asset('circular_dependency')

    def test_unicode_support(self):
        output = read(os.path.join(FIXTURES_DIR, 'unicode_support', 'output.js'))
        self.assertEqual(unicode(self.get_asset('unicode_support')), output)



import os
from .base import ExecCompiler


class CoffeeScriptCompiler(ExecCompiler):

    result_mimetype = 'application/javascript'
    executable = 'node'
    params = [os.path.join(os.path.dirname(__file__), 'coffeescript.js')]



import os
from .base import ExecCompiler


class StylusCompiler(ExecCompiler):

    result_mimetype = 'text/css'
    executable = 'node'
    params = [os.path.join(os.path.dirname(__file__), 'stylus.js')]

    def get_args(self):
        args = super(StylusCompiler, self).get_args()
        args.append(self.asset.absolute_path)
        return args



from gears.environment import PublicAssets
from unittest2 import TestCase


class PublicAssetsTests(TestCase):

    def setUp(self):
        self.public_assets = PublicAssets()

    def test_register_defaults(self):
        self.public_assets.register_defaults()
        self.assertEqual(self.public_assets, ['css/style.css', 'js/script.js'])

    def test_register(self):
        self.public_assets.register('js/app.js')
        self.public_assets.register('js/admin.js')
        self.assertEqual(self.public_assets, ['js/app.js', 'js/admin.js'])

    def test_register_twice(self):
        self.public_assets.register('js/app.js')
        self.public_assets.register('js/app.js')
        self.assertEqual(self.public_assets, ['js/app.js'])

    def test_unregister(self):
        self.public_assets.register('js/app.js')
        self.public_assets.unregister('js/app.js')
        self.assertEqual(self.public_assets, [])

    def test_unregister_if_does_not_exist(self):
        self.public_assets.unregister('js/admin.js')



import subprocess
from ..asset_handler import BaseAssetHandler


class CompressorFailed(Exception):
    pass


class BaseCompressor(BaseAssetHandler):
    pass


class ExecCompressor(BaseCompressor):

    executable = None
    params = []

    def __init__(self, executable=None):
        if executable is not None:
            self.executable = executable

    def __call__(self, asset):
        self.asset = asset
        p = subprocess.Popen(
            args=self.get_args(),
            stdin=subprocess.PIPE,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE)
        output, errors = p.communicate(input=self.asset.bundled_source.encode('utf-8'))
        if p.returncode != 0:
            raise CompressorFailed(errors)
        return output.decode('utf-8')

    def get_args(self):
        return [self.executable] + self.params



import re


class DirectivesParser(object):

    header_re = re.compile(r"""
        ^( \s* (
            ( /\* .*? \*/ ) |  # multiline comment
            ( // [^\n]* )+ |   # slash comment
            ( \# [^\n]* )+     # dash comment
        ) )+
    """, re.S | re.X)

    directive_re = re.compile(r"""
        ^ \s* (?:\*|//|\#) \s* = \s* ( \w+ [./'"\s\w-]* ) $
    """, re.X)

    def split_source(self, source):
        header_match = self.header_re.match(source)
        if not header_match:
            return '', source
        header = header_match.group(0)
        source = source[len(header):]
        return header, source

    def split_header(self, header):
        directives = []
        header_lines = []
        for line in header.splitlines():
            directive_match = self.directive_re.match(line)
            if directive_match:
                directives.append(directive_match.group(1))
            else:
                header_lines.append(line)
        return directives, '\n'.join(header_lines)

    def parse(self, source):
        header, source = self.split_source(source)
        directives, header = self.split_header(header)
        return directives, ''.join([header, source]).strip() + '\n'



import os


missing = object()


class cached_property(object):

    def __init__(self, func):
        self.func = func
        self.__name__ = func.__name__
        self.__doc__ = func.__doc__
        self.__module__ = func.__module__

    def __get__(self, obj, type=None):
        if obj is None:
            return self
        value = obj.__dict__.get(self.__name__, missing)
        if value is missing:
            value = self.func(obj)
            obj.__dict__[self.__name__] = value
        return value


def safe_join(base, *paths):
    if not os.path.isabs(base):
        raise ValueError("%r is not an absolute path." % base)
    base = os.path.normpath(base)
    path = os.path.normpath(os.path.join(base, *paths))
    if not path.startswith(base):
        raise ValueError("Path %r is outside of %r" % (path, base))
    return path


def unique(iterable, key=lambda x: x):
    yielded = set()
    for item in iterable:
        keyitem = key(item)
        if keyitem not in yielded:
            yielded.add(keyitem)
            yield item



from .coffeescript import CoffeeScriptCompiler
from .handlebars import HandlebarsCompiler
from .less import LessCompiler
from .sass import SASSCompiler
from .scss import SCSSCompiler
from .stylus import StylusCompiler



import os
from .base import ExecCompiler


class LessCompiler(ExecCompiler):

    result_mimetype = 'text/css'
    executable = 'node'
    params = [os.path.join(os.path.dirname(__file__), 'less.js')]



from __future__ import with_statement

import os
from gears.directives_parser import DirectivesParser
from unittest2 import TestCase


TESTS_DIR = os.path.dirname(__file__)
FIXTURES_DIR = os.path.join(TESTS_DIR, 'fixtures', 'directives_parser')


def read(file):
    with open(file) as f:
        return f.read()


class DirectivesParserTests(TestCase):

    def get_fixture_files(self, path):
        files = {}
        for filename in os.listdir(path):
            files[filename.split('.')[0]] = os.path.join(path, filename)
        return files['source'], files['output']

    def get_fixture(self, fixture):
        fixture_path = os.path.join(FIXTURES_DIR, fixture)
        return map(read, self.get_fixture_files(fixture_path))

    def check_asset(self, fixture, directives):
        source, output = self.get_fixture(fixture)
        result = DirectivesParser().parse(source)
        self.assertEqual(result[0], directives)
        self.assertEqual(result[1], output)

    def test_does_nothing_if_no_directives(self):
        self.check_asset('no_directives', [])

    def test_strips_whitespaces(self):
        self.check_asset('whitespaces', [])

    def test_multiline_comments(self):
        self.check_asset('multiline_comments', [
            'require reset',
            'require base',
        ])

    def test_handles_slash_comments(self):
        self.check_asset('slash_comments', [
            'require jquery',
            'require underscore'
        ])

    def test_handles_dash_comments(self):
        self.check_asset('dash_comments', [
            'require jquery',
            'require_tree .',
        ])

    def test_handle_multiple_comments(self):
        self.check_asset('multiple_comments', [
            'require jquery',
            'require underscore',
            'require backbone',
            'require models',
            'require collections',
            'require views',
        ])

    def test_skips_non_header_comments(self):
        self.check_asset('non_header_comments', ['require jquery'])


class DirectivesParserHeaderPatternTests(TestCase):

    def setUp(self):
        self.re = DirectivesParser().header_re

    def test_matches_multiline_comment(self):
        self.assertTrue(self.re.match('\n\n/*\nmultiline comment\n */'))

    def test_matches_multiple_miltiline_comments(self):
        self.assertTrue(self.re.match('/* comment */\n\n/* comment */'))

    def test_matches_slash_comment(self):
        self.assertTrue(self.re.match('  // this\n\n  // is\n  // comment'))

    def test_matches_dash_comment(self):
        self.assertTrue(self.re.match('  # this\n  \n  # is\n  # comment'))


class DirectivesParserDirectivePatternTests(TestCase):

    def setUp(self):
        self.re = DirectivesParser().directive_re

    def test_matches_directives(self):
        self.assertTrue(self.re.match(' *= require jquery'))
        self.assertTrue(self.re.match(' * =require jquery'))
        self.assertTrue(self.re.match('//= require jquery'))
        self.assertTrue(self.re.match(' #= require jquery'))
